package org.apidesign.apifest08.test;

import java.util.Currency;

import junit.framework.TestCase;

import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.ConvertorFactory;
import org.apidesign.apifest08.currency.MoneyImpl;

/** Finish the Convertor API, and then write bodies of methods inside
 * of this class to match the given tasks. To fullfil your task, use the
 * API define in the <code>org.apidesign.apifest08.currency</code> package.
 * Do not you reflection, or other hacks as your code
 * shall run without any runtime permissions.
 */
public class Task1Test extends TestCase {
	
	public static final Currency USD = Currency.getInstance("USD");
	public static final Currency CZK = Currency.getInstance("CZK");
	public static final Currency SKK = Currency.getInstance("SKK");
	
    public Task1Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    //
    // Imagine that there are three parts of the whole system:
    // 1. there is someone who knows the current exchange rate
    // 2. there is someone who wants to do the conversion
    // 3. there is the API between 1. and 2. which allows them to communicate
    // Please design such API
    //

    /** Create convertor that understands two currencies, CZK and
     *  USD. Make 1 USD == 17 CZK. This is a method provided for #1 group -
     *  e.g. those that know the exchange rate. They somehow need to create
     *  the objects from the API and tell them the exchange rate. The API itself
     *  knows nothing about any rates, before the createCZKtoUSD method is called.
     *
     * Creation of the convertor shall not require subclassing of any class
     * or interface on the client side.
     *
     * @return prepared convertor ready for converting USD to CZK and CZK to USD
     */
    public static Convertor createCZKtoUSD() {
    	return ConvertorFactory.createConvertor(new MoneyImpl(17,CZK), new MoneyImpl(1,USD));
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 80 CZK. Again this is method for the #1 group -
     *  it knows the exchange rate, and needs to use the API to create objects
     *  with the exchange rate. Anyone shall be ready to call this method without
     *  any other method being called previously. The API itself shall know
     *  nothing about any rates, before this method is called.
     *
     * Creation of the convertor shall not require subclassing of any class
     * or interface on the client side.
     * 
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK() {
    	return ConvertorFactory.createConvertor(new MoneyImpl(100,SKK), new MoneyImpl(80,CZK));
    }

    //
    // now the methods for group #2 follow:
    // this group knows nothing about exchange rates, but knows how to use
    // the API to do conversions. It somehow (by calling one of the factory
    // methods) gets objects from the API and uses them to do the conversions.
    //
    
    /** Use the convertor from <code>createCZKtoUSD</code> method and do few conversions
     * with it.
     */
    public void testCurrencyCZKUSD() throws Exception {
    	Convertor c = createCZKtoUSD();
        // convert $5 to CZK using c:
        assertEquals("Result is 85 CZK",new MoneyImpl(85,CZK), c.convert(new MoneyImpl(5,USD),CZK));

        // convert $8 to CZK
        assertEquals("Result is 136 CZK",new MoneyImpl(136,CZK), c.convert(new MoneyImpl(8,USD),CZK));

        // convert 1003CZK to USD
        assertEquals("Result is 59 USD", new MoneyImpl(59,USD), c.convert(new MoneyImpl(1003,CZK),USD));
        
    }

    /** Use the convertor from <code>createSKKtoCZK</code> method and do few conversions
     * with it.
     */
    public void testCurrencySKKCZK() throws Exception {
        Convertor c = createSKKtoCZK();
        // convert 16CZK using c:
        assertEquals("Result is 20 SKK", new MoneyImpl(20,SKK), c.convert(new MoneyImpl(16,CZK),SKK));

        // convert 500SKK to CZK
        assertEquals("Result is 400 CZK", new MoneyImpl(400,CZK), c.convert(new MoneyImpl(500,SKK),CZK));
    }

    /** Verify that the CZK to USD convertor knows nothing about SKK.
     */
    public void testCannotConvertToSKKwithCZKUSDConvertor() throws Exception {
        Convertor c = createCZKtoUSD();
        // convert $5 to SKK, the API shall say this is not possible
        try
        {
        	c.convert(new MoneyImpl(5, USD), SKK);
        	fail("Exception expected");
        }
        catch(IllegalArgumentException e)
        {
        	assertTrue("Ok",true);
        }
        // convert 500 SKK to CZK, the API shall say this is not possible
        try
        {
        	c.convert(new MoneyImpl(500, SKK), CZK);
        	fail("Exception expected");
        }
        catch(IllegalArgumentException e)
        {
        	assertTrue("Ok",true);
        }
    }

    /** Verify that the CZK to SKK convertor knows nothing about USD.
     */
    public void testCannotConvertToUSDwithCZKSKKConvertor() throws Exception {
        Convertor c = createSKKtoCZK();
        // convert $5 to SKK, the API shall say this is not possible
        try
        {
        	c.convert(new MoneyImpl(5, USD), SKK);
        	fail("Exception expected");
        }
        catch(IllegalArgumentException e)
        {
        	assertTrue("Ok",true);
        }
        // convert 500 CZK to USD, the API shall say this is not possible
        try
        {
        	c.convert(new MoneyImpl(500, CZK), USD);
        	fail("Exception expected");
        }
        catch(IllegalArgumentException e)
        {
        	assertTrue("Ok",true);
        }
    }
}
