package org.apidesign.apifest08.test;

import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.CurrencyValue;
import org.apidesign.apifest08.currency.ExchangeRateDataSource;
import org.apidesign.apifest08.currency.ExchangeRateProvider;
import org.apidesign.apifest08.currency.ExchangeRateProvider.ExchangeRateRequest;
import org.apidesign.apifest08.currency.ExchangeRateProvider.ExchangeRateResponse;
import org.apidesign.apifest08.currency.ExchangeRateValue;

/** The exchange rates are not always the same. They are changing. Day by day,
 * hour by hour, minute by minute. For every bank it is important to always
 * have the actual exchange rate available in the system. That is why let's
 * create a pluggable convertor that will always have up to date value of its
 * exchange rate.
 * <p>
 * The quest for today is to allow 3rd party developer to write a convertor
 * that adjusts its exchange rate everytime it is queried. This convertor is
 * written by independent vendor, the vendor knows only your Convertor API,
 * he does not know how the whole system looks and how the convertor is supposed
 * to be used.
 */
public class Task3Test extends TestCase {

    public Task3Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //
    /** Without knowing anything about the surrounding system, write an
     * implementation of convertor that will return different rates everytime
     * it is queried. Convert USD to CZK and vice versa. Start with the rate of
     * 1USD = 16CZK and adjust it in favor of CZK by 0.01 CZK with every query.
     * As soon as you reach 1USD = 15CZK adjust it by 0.01 CZK in favor of USD
     * until you reach 1USD = 16CZK
     *
     * @return new instance of "online" USD and CZK convertor starting with rate 1USD = 16CZK
     */
    public static Convertor<Double, String> createOnlineCZKUSDConvertor() {
        // initial rate: 1USD = 16CZK
        // 2nd query 1USD = 15.99CZK
        // 3rd query 1USD = 15.98CZK
        // until 1USD = 15.00CZK
        // then 1USD = 15.01CZK
        // then 1USD = 15.02CZK
        // and so on and on up to 1USD = 16CZK
        // and then another round to 15, etc.
        return Convertor.getConvertorDataSourceDoubleString(
                ExchangeRateDataSource.getExchangeRateDataSource(
                "USD", "CZK", new ExchangeRateProvider<Double, String>() {

            double currentRate = 16d;
            double step;

            public void getExchangeRate(ExchangeRateRequest<Double, String> request,
                    ExchangeRateResponse<Double, String> response) {
                if ((request.getCurrencyAIdentifier().equals("CZK") && request.getCurrencyBIdentifier().equals("USD")) ||
                        (request.getCurrencyAIdentifier().equals("USD") && request.getCurrencyBIdentifier().equals("CZK"))) {
                    response.setExchangeRate(ExchangeRateValue.getExchangeRate(
                            CurrencyValue.getCurrencyValue(1d, "USD"),
                            CurrencyValue.getCurrencyValue(currentRate, "CZK")));

                    if (currentRate == 16d) {
                        step = -0.01;
                    }
                    if (currentRate == 15d) {
                        step = 0.01;
                    }
                    currentRate += step;
                } else {
                    throw new IllegalArgumentException("No exchange rate for requested currencies!");
                }
            }
        }));
    }

    public void testFewQueriesForOnlineConvertor() {
        Convertor<Double, String> c = createOnlineCZKUSDConvertor();
        doFewQueriesForOnlineConvertor(c);
    }

    static void doFewQueriesForOnlineConvertor(Convertor<Double, String> c) {
        CurrencyValue<Double, String> result;
        // convert $5 to CZK using c:
        //assertEquals("Result is 80 CZK");
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(5d, "USD"));
        assertEquals(CurrencyValue.getCurrencyValue(80d, "CZK"), result);

        // convert $8 to CZK using c:
        //assertEquals("Result is 127.92 CZK");
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(8d, "USD"));
        assertEquals(CurrencyValue.getCurrencyValue(127.92d, "CZK"), result);

        // convert $1 to CZK using c:
        //assertEquals("Result is 15.98 CZK");
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(1d, "USD"));
        assertEquals(CurrencyValue.getCurrencyValue(15.98d, "CZK"), result);

        // convert 15.97CZK to USD using c:
        //assertEquals("Result is 1$");
        result = c.convert("USD", CurrencyValue.getCurrencyValue(15.97d, "CZK"));
        assertEquals(CurrencyValue.getCurrencyValue(1d, "USD"), result);
    }

    /** Join the convertors and show they behave sane.
     */
    public void testOnlineConvertorComposition() throws Exception {
        Convertor<Double, String> c = Task2Test.merge(
                createOnlineCZKUSDConvertor(),
                Task1Test.createSKKtoCZK());

        CurrencyValue<Double, String> result;
        // convert 16CZK to SKK using c:
        // assertEquals("Result is 20 SKK");
        result = c.convert("SKK", CurrencyValue.getCurrencyValue(16d, "CZK"));
        assertEquals(CurrencyValue.getCurrencyValue(20d, "SKK"), result);

        // convert 500SKK to CZK using c:
        // assertEquals("Result is 400 CZK");
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(500d, "SKK"));
        assertEquals(CurrencyValue.getCurrencyValue(400d, "CZK"), result);

        doFewQueriesForOnlineConvertor(c);
    }
}
